from daydream.plugins.aws.nodes.base import AwsNode


class AwsAlbTargetGroup(AwsNode):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["TargetGroupArn"],
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            *((lb, "routes_requests_for") for lb in self.raw_data.get("LoadBalancerArns", [])),
            self.raw_data.get("VpcId"),
        ]



import asyncio
import gzip
import json
import os
import re
import secrets
from datetime import UTC, date, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Literal, cast

from pydantic import AwareDatetime, BaseModel
from sh import Command, ErrorReturnCode

from daydream.lib.log_anomalies import LogLine
from daydream.plugins.aptible.nodes.aptible_aws_instance import AptibleAwsInstance
from daydream.plugins.aptible.nodes.base import AptibleNode
from daydream.utils import print

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from daydream.plugins.aptible.nodes.aptible_container import AptibleContainer


LOGS_DIR = Path.home() / ".daydream" / "logs"
LOG_FILENAME_PATTERN = re.compile(r"^.*\.archived\.gz$")
LOG_FILENAME_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
MAX_CONTAINER_LOG_CHECKS = 2

LogResourceType = Literal["service", "container"]

random = secrets.SystemRandom()


class LogBucketConfig(BaseModel):
    bucket_name: str
    region: str
    decryption_keys: str


async def get_logs_for_time_range(
    stack: str,
    node: AptibleNode,
    containers: list["AptibleContainer"],
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
) -> list[LogLine]:
    start_date = start_dt.date()
    end_date = end_dt.date()

    log_dir = LOGS_DIR / f"{stack}-{node.raw_data['_type']}-{node.node_id}-{start_date}-{end_date}"
    log_dir.mkdir(parents=True, exist_ok=True)
    log_filenames = log_files_in_dir(log_dir, start_dt, end_dt)

    if not log_filenames:
        print(
            f"Downloading logs for {stack} {node.raw_data['_type']} {node.node_id} {start_date} {end_date}"
        )

        if len(containers) > MAX_CONTAINER_LOG_CHECKS:
            print(f"Sampling {len(containers)} containers to {MAX_CONTAINER_LOG_CHECKS}")
            containers = random.sample(containers, MAX_CONTAINER_LOG_CHECKS)

        bucket_config = await get_log_bucket_config(stack)
        await asyncio.gather(
            download_container_logs(stack, containers, start_dt, end_dt, log_dir),
            download_s3_logs(stack, containers, start_date, end_date, bucket_config, log_dir),
        )

        log_filenames = log_files_in_dir(log_dir, start_dt, end_dt)
    else:
        print(
            f"Using cached logs for {stack} {node.raw_data['_type']} {node.node_id} {start_date} {end_date}"
        )

    return await get_log_lines(log_filenames, start_dt, end_dt)


async def get_log_lines(
    log_filenames: list[Path], start_dt: datetime, end_dt: datetime
) -> list[LogLine]:
    log_lines: list[LogLine] = []
    print(f"Getting log lines for {len(log_filenames)} files")

    if len(log_filenames) == 0:
        return log_lines

    # Ensure input datetimes are timezone-aware
    if start_dt.tzinfo is None:
        start_dt = start_dt.replace(tzinfo=UTC)
    if end_dt.tzinfo is None:
        end_dt = end_dt.replace(tzinfo=UTC)

    # Get filesize of all log files
    log_file_size = sum([f.stat().st_size for f in log_filenames])
    print(f"Total log file size: {log_file_size}")

    if log_file_size == 0:
        return log_lines

    # Calculate sampling ratio based on a target filesize
    target_filesize = 100 * 1024 * 1024  # 100MB
    sampling_ratio = min(1.0, target_filesize / log_file_size)
    print(f"Sampling ratio: {sampling_ratio}")

    for log_filename in log_filenames:
        with gzip.open(log_filename, "rt") as f:
            for line in f:
                if random.random() < sampling_ratio:
                    try:
                        json_line = json.loads(line)
                        dt = parse_log_line_timestamp(json_line["time"])
                        if start_dt <= dt <= end_dt:
                            log_lines.append((dt, json_line["log"].strip()))
                    except (json.JSONDecodeError, ValueError):
                        pass

    return log_lines


def parse_log_line_timestamp(ts: str) -> datetime:
    ts_parts = ts.split(".")

    if len(ts_parts) == 2:
        base, nanos = ts_parts
        micros = (nanos[:-1] + "000000")[:6]  # Remove Z, pad/truncate to 6 digits
        ts_fixed = f"{base}.{micros}Z"
        dt = datetime.strptime(ts_fixed, "%Y-%m-%dT%H:%M:%S.%fZ")
    else:
        dt = datetime.strptime(ts, "%Y-%m-%dT%H:%M:%SZ")

    # Make the datetime timezone-aware by assuming UTC
    return dt.replace(tzinfo=UTC)


async def get_log_bucket_config(stack: str) -> LogBucketConfig:
    log_bucket = await get_pancake_setting(stack, "sweetness/nonsensitive/DOCKER_LOGS_BUCKET_NAME")
    bucket_region = log_bucket.replace("aptible-docker-logs-", "")
    log_encryption_keys = await get_pancake_setting(
        stack, "sweetness/sensitive/LOG_ENCRYPTION_KEYS"
    )
    return LogBucketConfig(
        bucket_name=log_bucket,
        region=bucket_region,
        decryption_keys=log_encryption_keys,
    )


async def download_s3_logs(
    stack: str,
    containers: list["AptibleContainer"],
    start_date: date,
    end_date: date,
    bucket_config: LogBucketConfig,
    download_dir: Path,
) -> None:
    for container in containers:
        print(f"Downloading s3 logs for {container.raw_data['docker_name']}")

        try:
            aptible_cli = Command("aptible")
            await aptible_cli.logs_from_archive(
                stack=stack,
                container_id=container.raw_data["docker_name"],
                start_date=start_date.strftime("%Y-%m-%d"),
                end_date=end_date.strftime("%Y-%m-%d"),
                region=bucket_config.region,
                bucket=bucket_config.bucket_name,
                decryption_keys=bucket_config.decryption_keys,
                download_location=str(download_dir),
                _async=True,
            )
        except ErrorReturnCode as e:
            print(f"Error running command: {e}")


async def download_container_logs(
    stack: str,
    containers: list["AptibleContainer"],
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
    download_dir: Path,
) -> None:
    pancake = cast("Callable[..., Awaitable[str]]", Command(os.getenv("PANCAKE_PATH")))

    log_lines: list[LogLine] = []

    pancake_env = os.environ.copy()
    pancake_env["THOR_SILENCE_DEPRECATION"] = "1"

    for container in containers:
        if container.raw_data["deleted_at"]:
            continue

        try:
            aws_instance = await anext(container.iter_neighboring(AptibleAwsInstance))
        except StopAsyncIteration as e:
            raise ValueError("Container has no parent AWS instance") from e

        container_id = container.raw_data["docker_name"]
        instance_name = aws_instance.raw_data["runtime_data"]["hostname"]

        print(f"Downloading container logs for {instance_name} {container_id}")
        output_lines = []

        try:
            output = await pancake(
                "stack:ssh",
                stack,
                "--instance",
                instance_name,
                "sudo",
                "docker",
                "container",
                "logs",
                container_id,
                "--timestamps",
                "--since",
                start_dt.strftime("%Y-%m-%dT%H:%M:%SZ"),
                "--until",
                end_dt.strftime("%Y-%m-%dT%H:%M:%SZ"),
                _async=True,
                _env=pancake_env,
            )
            output_lines = output.splitlines()
            print(f"Found {len(output_lines)} log lines for {container_id}")
        except ErrorReturnCode as e:
            print(f"Error running command: {e}")

        for line in output_lines:
            try:
                dt, msg = line.split(" ", maxsplit=1)
                dt = parse_log_line_timestamp(dt)
                log_lines.append((dt, msg))
            except ValueError:
                pass

    # If we found any log lines, write them to a gzipped jsonl file (same format
    # as s3 logs)
    if len(log_lines) > 0:
        with gzip.open(
            download_dir
            / f"?????????-json.log.{start_dt.strftime(LOG_FILENAME_DATETIME_FORMAT)}.{end_dt.strftime(LOG_FILENAME_DATETIME_FORMAT)}.archived.gz",
            "wt",
        ) as f:
            for dt, msg in log_lines:
                log_line_json = {
                    "time": dt.isoformat(),
                    "log": msg,
                }
                f.write(json.dumps(log_line_json) + "\n")


def log_files_in_dir(log_dir: Path, start_dt: datetime, end_dt: datetime) -> list[Path]:
    filenames = [
        f for f in log_dir.glob("**/*") if f.is_file() and LOG_FILENAME_PATTERN.match(f.name)
    ]
    matching_files = []
    for filename in filenames:
        dt_range = str(filename).split("json.log.")[1].replace(".archived.gz", "")
        dt_range = dt_range.removeprefix("1.")

        log_start, log_end = dt_range.split(".")
        log_start_dt = datetime.strptime(log_start, LOG_FILENAME_DATETIME_FORMAT)
        log_start_dt = log_start_dt.replace(tzinfo=UTC)
        log_end_dt = datetime.strptime(log_end, LOG_FILENAME_DATETIME_FORMAT)
        log_end_dt = log_end_dt.replace(tzinfo=UTC)

        # Check if the log file's time range overlaps with our target period
        if (
            (start_dt <= log_start_dt <= end_dt)
            or (start_dt <= log_end_dt <= end_dt)
            or (log_start_dt <= start_dt <= log_end_dt)
        ):
            print(f"{filename} {log_start_dt} {log_end_dt}")
            matching_files.append(filename)

    return matching_files


async def get_pancake_setting(stack: str, setting: str) -> str:
    pancake = cast("Callable[..., Awaitable[str]]", Command(os.getenv("PANCAKE_PATH")))

    try:
        result = await pancake(
            "stack:settings:view_single",
            stack,
            setting,
            _async=True,
        )
        return result.strip()
    except ErrorReturnCode as e:
        print(f"Error running command: {e}")

    return ""



import anyio
import typer

from daydream import mcp
from daydream.cli._app import app
from daydream.cli.options import PROFILE_OPTION
from daydream.telemetry import client as telemetry


@app.command()
def start(
    profile: str = PROFILE_OPTION,
    disable_sse: bool = typer.Option(
        False, "--disable-sse", help="Disable the SSE transport for the MCP Server"
    ),
    disable_stdio: bool = typer.Option(
        False, "--disable-stdio", help="Disable the stdio transport for the MCP Server"
    ),
) -> None:
    """Start the Daydream MCP Server"""

    async def _start() -> None:
        await telemetry.send_event(
            {
                "command": "start",
                "profile": profile,
                "disable_sse": disable_sse,
                "disable_stdio": disable_stdio,
            }
        )
        await mcp.start(profile, disable_sse, disable_stdio)

    anyio.run(_start)



from collections import defaultdict
from collections.abc import AsyncIterator
from pathlib import Path
from typing import TypeVar, cast

import networkx as nx
from anyio import Lock
from pydantic_core import from_json, to_json, to_jsonable_python

from daydream.utils import generate_contrasting_colors, print, strip_secrets

from .edges import Edge
from .nodes import NODE_REGISTRY, Node

_NodeType = TypeVar("_NodeType", bound="Node")


class Graph:
    _lock: Lock
    _identifier_mapping: defaultdict[str, set[str]]

    def __init__(self, path: Path | str | None = None) -> None:
        self._lock = Lock()
        self._identifier_mapping = defaultdict(set)
        self._path = Path(path) if path else None
        self._loaded = False

    @property
    def digraph(self) -> nx.DiGraph:
        if not self._loaded:
            if self._path:
                if not self._path.exists():
                    self._digraph = nx.DiGraph()
                    self._path.write_bytes(
                        to_json(
                            nx.node_link_data(
                                self._digraph,
                                edges="edges",
                            ),
                            indent=2,
                        )
                    )

                json = from_json(self._path.read_bytes())
                self._digraph = nx.node_link_graph(json, edges="edges")
            else:
                self._digraph = nx.DiGraph()

            # Convert the nodes to the proper Node objects.
            for _nid, node in self._digraph.nodes(data=True):
                node_data = node.pop("data")
                node_key = ":".join([node_data["node_source"], node_data["node_type"]])
                node["data"] = NODE_REGISTRY[node_key](
                    **node_data,
                    _graph=self,
                )

            self._loaded = True

        return self._digraph

    async def add_node(self, node: Node) -> None:
        """Add a node to the graph."""
        async with self._lock:
            self.digraph.add_node(node.nid, data=node)

            for identifier in (node.nid, *await node.get_identifiers()):
                if not identifier:
                    continue
                self._identifier_mapping[identifier].add(node.nid)

    async def get_node(self, nid: str) -> Node:
        """Get a node from the graph."""
        try:
            node = self.digraph.nodes[nid]
        except KeyError as ex:
            raise LookupError(f"Node {nid} not found in graph") from ex
        return node["data"]

    async def iter_nodes(self) -> AsyncIterator[Node]:
        """Get all nodes in the graph."""
        for _, node in self.digraph.nodes(data=True):
            yield cast("Node", node["data"])

    async def iter_predecessors(
        self,
        node: Node,
    ) -> AsyncIterator[Node]:
        """Get all predecessors of a node."""
        for predecessor_nid in self.digraph.predecessors(node.nid):
            yield self.digraph.nodes[predecessor_nid]["data"]

    async def iter_successors(self, node: Node) -> AsyncIterator[Node]:
        """Get all successors of a node."""
        for successor_nid in self.digraph.successors(node.nid):
            yield self.digraph.nodes[successor_nid]["data"]

    async def iter_neighbors(self, node: Node) -> AsyncIterator[Node]:
        """Get all neighbors of a node."""
        async for neighbor in self.iter_predecessors(node):
            yield neighbor

        async for neighbor in self.iter_successors(node):
            yield neighbor

    async def iter_neighboring(
        self,
        node: Node,
        only_type: type[_NodeType] | tuple[type[_NodeType], ...],
    ) -> AsyncIterator[_NodeType]:
        """Get all neighbors of a node that are of a specific type."""
        only_types = only_type if isinstance(only_type, tuple) else (only_type,)

        async for neighbor in self.iter_neighbors(node):
            if not isinstance(neighbor, only_types):
                continue
            yield cast("_NodeType", neighbor)

    async def iter_neighborhood_edges(self, node: Node, max_depth: int = 1) -> AsyncIterator[Edge]:
        """Get all edges in the neighborhood of a node."""
        for reverse in (True, False):
            for source_nid, destination_nid in nx.bfs_edges(
                self._digraph, node.nid, depth_limit=max_depth, reverse=reverse
            ):
                source_node = self.digraph.nodes[source_nid]["data"]
                destination_node = self.digraph.nodes[destination_nid]["data"]
                properties = (
                    self.digraph.edges[source_nid, destination_nid]
                    if not reverse
                    else self.digraph.edges[destination_nid, source_nid]
                )

                yield Edge(
                    source_node=source_node,
                    destination_node=destination_node,
                    properties=properties,
                )

    async def add_edge(self, edge: Edge) -> None:
        """Add an edge between nodes to the graph."""
        async with self._lock:
            self.digraph.add_node(edge.source_node.nid, data=edge.source_node)
            self.digraph.add_node(edge.destination_node.nid, data=edge.destination_node)

            edge_properties = edge.properties
            if "relationship_type" in edge_properties:
                edge_properties["label"] = edge_properties["relationship_type"]

            self.digraph.add_edge(
                edge.source_node.nid,
                edge.destination_node.nid,
                **edge_properties,
            )

    async def iter_edges(self) -> AsyncIterator[Edge]:
        """Get all edges in the graph."""
        for source_nid, destination_nid, properties in self.digraph.edges(data=True):
            source_node = self.digraph.nodes[source_nid]["data"]
            destination_node = self.digraph.nodes[destination_nid]["data"]

            yield Edge(
                source_node=source_node,
                destination_node=destination_node,
                properties=properties,
            )

    async def infer_edges(self) -> None:
        """Infer edges between nodes based on the identifier mapping."""
        edge_count = self.digraph.number_of_edges()

        async for node in self.iter_nodes():
            for ref in await node.get_reference_identifiers():
                if isinstance(ref, tuple) and len(ref) == 2:
                    reference_identifier, relationship_type = ref
                else:
                    reference_identifier = ref
                    relationship_type = "related_to"

                if not reference_identifier:
                    continue

                related_node_ids = self._identifier_mapping.get(reference_identifier, set())

                if not related_node_ids:
                    continue

                if len(related_node_ids) > 1:
                    # TODO: Handle multiple related nodes (i.e. ambiguous edges).
                    continue

                related_node_id = next(iter(related_node_ids))
                related_node = await self.get_node(related_node_id)

                print(f"Inferred edge between {node.nid} and {related_node.nid}")

                await self.add_edge(
                    Edge(
                        source_node=node,
                        destination_node=related_node,
                        properties={
                            "relationship_type": relationship_type,
                        },
                    )
                )

        print(f"Inferred {self.digraph.number_of_edges() - edge_count} edges")

    async def get_topology(self) -> "Graph":
        topology = Graph()

        async for edge in self.iter_edges():
            source_node = edge.source_node.model_copy(update={"node_id": ""})
            destination_node = edge.destination_node.model_copy(update={"node_id": ""})

            # Add nodes (types) and edge between them
            await topology.add_edge(
                Edge(
                    source_node=source_node,
                    destination_node=destination_node,
                    properties={"label": edge.properties["relationship_type"]},
                )
            )

        return topology

    async def to_pydot(self) -> str:
        """Convert the graph to a Pydot graph."""
        node_types = {f"{node.node_source}:{node.node_type}" async for node in self.iter_nodes()}
        node_colors = dict(
            zip(node_types, generate_contrasting_colors(len(node_types)), strict=True)
        )

        graph: nx.DiGraph = nx.relabel_nodes(
            self._digraph,
            # The mapping can be a callable, despite the type hint.
            #
            # We also quote the labels, since identifiers can often contain
            # colons, which are special in Pydot.
            mapping=lambda old: f'"{old}"',  # type: ignore
            copy=True,
        )

        # Set the defaults.
        graph.graph["graph"] = {"nodesep": 1}
        graph.graph["node"] = {
            "shape": "box",
            "style": "rounded,filled",
            "fontname": "Helvetica",
            "margin": "0.25",
        }

        # Set the node colors.
        for _, node in graph.nodes(data=True):
            node_data = cast("Node", node["data"])
            node_key = f"{node_data.node_source}:{node_data.node_type}"
            fill_color, font_color = node_colors[node_key]
            node["fillcolor"] = fill_color
            node["fontcolor"] = font_color
            del node["data"]

        return nx.nx_pydot.to_pydot(graph)

    async def save(self, path: Path | str) -> None:
        """Save the graph to a file."""
        async with self._lock:
            data = nx.node_link_data(self._digraph, edges="edges")
            print("Scrubbing secrets from graph...")
            cleaned_data = strip_secrets(to_jsonable_python(data))
            path = Path(path)
            path.write_bytes(
                to_json(
                    {
                        **cleaned_data,
                        "_identifier_mapping": dict(self._identifier_mapping),
                    },
                    indent=2,
                )
            )
